from pathlib import Path

import pkg_resources


def get_path_of_data_dir() -> Path:
    """
    get the path of the package data directory

    :returns:

    """
    file_path: str = pkg_resources.resource_filename("ronswanson", "data")

    return Path(file_path)


def get_path_of_data_file(data_file: str) -> Path:
    """
    get the path of a dat file

    :param data_file: name of the data file
    :type data_file: str
    :returns:

    """
    file_path: Path = get_path_of_data_dir() / data_file

    return file_path


# def get_path_of_user_config() -> Path:
#     """
#     get the path to the user configuration

#     :returns:

#     """
#     config_path: Path = Path().home() / ".config" / "ronswanson"

#     if not config_path.exists():

#         config_path.mkdir(parents=True)

#     return config_path


__all__ = [
    "get_path_of_data_file",
    "get_path_of_data_dir",
]



import logging

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from .configuration import ronswanson_config


class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


_console_formatter = logging.Formatter(
    ' %(message)s',
    datefmt="%H:%M:%S",
)


_theme = {}

# Banner
_theme["h1"] = "deep_sky_blue3"
_theme["status.spinner"] = "cyan2"
_theme["status.text"] = "deep_sky_blue4"
_theme["repr.filename"] = "blue"
_theme["repr.number"] = "white"
_theme["repr.path"] = "grey37"
_theme["repr.str"] = "grey37"
_theme["repr.tag_name"] = "white"
_theme["repr.url"] = "not bold not italic underline grey84"
_theme["log.time"] = "green1"
_theme["log.message"] = "bold grey78"
_theme["logging.level.debug"] = "blue_violet"
_theme["logging.level.error"] = "blink bold bright_red"
_theme["logging.level.info"] = "medium_spring_green"
_theme["logging.level.warning"] = "blink medium_orchid"


mytheme = Theme(_theme)


console = Console(theme=mytheme)

ronswanson_console_log_handler = RichHandler(
    level=ronswanson_config.logging.level,
    rich_tracebacks=True,
    markup=True,
    console=console,
)
ronswanson_console_log_handler.setFormatter(_console_formatter)

warning_filter = LogFilter(logging.WARNING)


def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    ronswanson_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    ronswanson_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    ronswanson_console_log_handler.setLevel(level)


def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(logging.DEBUG)

    if ronswanson_config.logging.on:

        # add the handlers
        log.addHandler(ronswanson_console_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



import itertools
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

import numpy as np
import yaml

from .utils.cartesian_product import cartesian_jit
from .utils.logging import setup_logger

log = setup_logger(__name__)


@dataclass(frozen=True)
class EnergyGrid:

    vmin: Optional[float] = None
    vmax: Optional[float] = None
    scale: Optional[str] = None
    n_points: Optional[int] = None
    values: Optional[np.ndarray] = None
    custom: bool = False

    def __post_init__(self):

        if not self.custom:

            # we will build a grid

            if (self.vmin is None) or (self.vmax is None):

                log.error("non-custom grids must include vmin and vmax")

                raise AssertionError

            if self.scale is None:

                log.error(
                    "non-custom grids must include scale 'log' or 'linear'"
                )

                raise AssertionError

            else:

                if self.scale not in ['log', 'linear']:

                    log.error(
                        "non-custom grids must include scale 'log' or 'linear'"
                    )

                    raise AssertionError

            if self.n_points is None:

                log.error("non-custom grids must include n_points")

                raise AssertionError

        else:

            if self.values is None:

                log.error("custom grids must include values")

                raise AssertionError

    @property
    def grid(self) -> np.ndarray:

        if self.custom:

            return self.values

        else:

            if self.scale.lower() == 'log':

                return np.geomspace(self.vmin, self.vmax, self.n_points)

            else:

                return np.linspace(self.vmin, self.vmax, self.n_points)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Parameter":

        inputs = {}
        inputs["custom"] = d["custom"]

        if d["custom"]:

            inputs["values"] = np.array(d["values"])

        else:

            inputs["vmin"] = float(d["vmin"])
            inputs["vmax"] = float(d["vmax"])
            inputs["scale"] = d["scale"]
            inputs["n_points"] = d["n_points"]

        return cls(**inputs)

    def to_dict(self) -> Dict[str, Any]:

        out = dict(custom=self.custom)

        if self.custom:

            out["values"] = self.values.tolist()

        else:

            out["vmin"] = self.vmin
            out["vmax"] = self.vmax
            out["scale"] = self.scale
            out["n_points"] = self.n_points

        return out


_invalid_parameter_names = ("redshift", "x", "y", "z", "K", "scale")


@dataclass(frozen=True)
class Parameter:

    name: str
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    scale: Optional[str] = None
    n_points: Optional[int] = None
    values: Optional[np.ndarray] = None
    custom: bool = False
    #   grid: np.ndarray = field(init=False)

    def __post_init__(self):

        if self.name.lower() in _invalid_parameter_names:

            msg = f"the parameter name: {self.name} is protected and cannot be used"

            log.error(
                f"the parameter name: {self.name} is protected and cannot be used"
            )

            raise RuntimeError(msg)

        if not self.custom:

            # we will build a grid

            if (self.vmin is None) or (self.vmax is None):

                log.error("non-custom grids must include vmin and vmax")

                raise AssertionError

            if self.scale is None:

                log.error(
                    "non-custom grids must include scale 'log' or 'linear'"
                )

                raise AssertionError

            else:

                if self.scale not in ['log', 'linear']:

                    log.error(
                        "non-custom grids must include scale 'log' or 'linear'"
                    )

                    raise AssertionError

            if self.n_points is None:

                log.error("non-custom grids must include n_points")

                raise AssertionError

        else:

            if self.values is None:

                log.error("custom grids must include values")

                raise AssertionError

    @property
    def grid(self) -> np.ndarray:

        if self.custom:

            return self.values

        else:

            if self.scale.lower() == 'log':

                return np.geomspace(self.vmin, self.vmax, self.n_points)

            else:

                return np.linspace(self.vmin, self.vmax, self.n_points)

    @classmethod
    def from_dict(cls, name: str, d: Dict[str, Any]) -> "Parameter":

        inputs = {}
        inputs["custom"] = d["custom"]

        if d["custom"]:

            inputs["values"] = np.array(d["values"])

        else:

            inputs["vmin"] = d["vmin"]
            inputs["vmax"] = d["vmax"]
            inputs["scale"] = d["scale"]
            inputs["n_points"] = d["n_points"]

        return cls(name, **inputs)

    def to_dict(self) -> Dict[str, Any]:

        out = dict(custom=self.custom)

        if self.custom:

            out["values"] = self.values.tolist()

        else:

            out["vmin"] = self.vmin
            out["vmax"] = self.vmax
            out["scale"] = self.scale
            out["n_points"] = self.n_points

        return out


@dataclass(frozen=False)
class ParameterGrid:

    parameter_list: List[Parameter]
    energy_grid: List[EnergyGrid]
    full_grid: Optional[np.ndarray] = field(init=False)

    def __post_init__(self):

        object.__setattr__(
            self,
            'full_grid',
            cartesian_jit([p.grid for p in self.parameter_list]),
        )

    @property
    def n_points(self) -> int:

        n = 1
        for param in self.parameter_list:

            n *= len(param.grid)

        return n

    @property
    def n_parameters(self) -> int:

        return len(self.parameter_list)

    @classmethod
    def from_dict(cls, d: Dict[str, Dict[str, Any]]) -> "ParameterGrid":

        # make sure to sort so that we always have the same parameter
        # ordering

        is_multi_output: bool = False
        n_energy_grids = 1

        for k in d.keys():

            if "energy_grid" in k:

                if len(k.split("_")) == 3:

                    if is_multi_output:

                        # we have already detected one

                        n_energy_grids += 1

                    is_multi_output = True

        if not is_multi_output:

            energy_grid = [EnergyGrid.from_dict(d.pop("energy_grid"))]

        else:

            energy_grid = [
                EnergyGrid.from_dict(d.pop(f"energy_grid_{i}"))
                for i in range(n_energy_grids)
            ]

        pars = list(d.keys())

        pars.sort()

        par_list = [
            Parameter.from_dict(par_name, d[par_name]) for par_name in pars
        ]

        return cls(par_list, energy_grid)

    @classmethod
    def from_yaml(cls, file_name: str) -> "ParameterGrid":

        with open(file_name, 'r') as f:

            inputs = yaml.load(stream=f, Loader=yaml.SafeLoader)

        return cls.from_dict(inputs)

    @property
    def parameter_names(self) -> List[str]:

        return [p.name for p in self.parameter_list]

    def to_dict(self) -> Dict[str, Dict[str, Any]]:

        out = {}

        for p in self.parameter_list:

            out[p.name] = p.to_dict()

        if len(self.energy_grid) == 1:

            out['energy_grid'] = self.energy_grid[0].to_dict()

        else:

            for i, eg in enumerate(self.energy_grid):

                out[f"energy_grid_{i}"] = eg.to_dict()

        return out

    def write(self, file_name: str) -> None:

        with open(file_name, "w") as f:

            yaml.dump(
                stream=f,
                data=self.to_dict(),
                default_flow_style=False,
                Dumper=yaml.SafeDumper,
            )

    def at_index(
        self, i: int, as_array: bool = False
    ) -> Union[Dict[str, float], np.ndarray]:
        """
        return the ith set of parameters

        :param i:
        :type i: int
        :returns:

        """

        result = self.full_grid[i]

        if not as_array:

            d = OrderedDict()

            for k, v in zip(self.parameter_names, result):

                d[k] = v

            return d

        else:

            return result



import time
from contextlib import contextmanager
from pathlib import Path

import h5py
import numpy as np

from .logging import setup_logger
from .configuration import ronswanson_config

log = setup_logger(__name__)


@contextmanager
def open_component_file(database_file_name: str, sim_id: int) -> h5py.File:

    p = Path(database_file_name)

    if ronswanson_config.slurm.store_dir is None:

        parent_dir = p.absolute().parent

    else:

        parent_dir = Path(ronswanson_config.slurm.store_dir).absolute()

    multi_file_dir: Path = parent_dir / Path(f"{p.stem}_store")

    if not multi_file_dir.exists():

        log.debug(f"creating the store directory: {multi_file_dir}")

        multi_file_dir.mkdir()

    this_file: Path = multi_file_dir / f"sim_store_{sim_id}.h5"

    if this_file.exists():

        log.error(f"{this_file} is getting ready to be written but it exists!")

        raise RuntimeError(
            f"{this_file} is getting ready to be written but it exists!"
        )

    f = h5py.File(str(this_file), "w")

    try:

        yield f

    finally:

        f.close()

        log.debug(f"closing {this_file}")



import re
import time
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional

import h5py
import numpy as np
from tqdm.auto import tqdm

import ronswanson.grids as grids
from ronswanson.utils.color import Colors
from ronswanson.utils.configuration import ronswanson_config

from .utils.file_open import open_component_file
from .utils.logging import setup_logger

log = setup_logger(__name__)


class Timer(object):
    """
    A timing context manager
    """

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.total_time = time.perf_counter() - self.start_time


class Simulation(metaclass=ABCMeta):
    def __init__(
        self,
        simulation_id: int,
        parameter_set: Dict[str, float],
        energy_grid: List[grids.EnergyGrid],
        out_file: str,
        num_outputs: int = 1,
        num_meta_parameters: Optional[int] = None,
    ) -> None:

        """
        Generic simulation class

        :param simulation_id:
        :type simulation_id: int
        :param parameter_set:
        :type parameter_set: Dict[str, float]
        :param energy_grid:
        :type energy_grid: np.ndarray
        :param out_file:
        :type out_file: str
        :returns:
        """
        self._out_file: str = out_file
        self._parameter_set: Dict[str, float] = parameter_set
        self._simulation_id: int = simulation_id
        self._energy_grid: List[grids.EnergyGrid] = energy_grid
        self._num_outputs: int = num_outputs
        self._num_meta_parameters: Optional[int] = num_meta_parameters

        if not len(self._energy_grid) == self._num_outputs:

            log.error(f"requested number of outputs {self._num_outputs}")
            log.error(f"but only have {len(self._energy_grid)} energy grids")

            raise RuntimeError()

    def run(self) -> None:

        """
        run this simulation
        :returns:

        """

        # run the simulation

        log.debug(f"simulation {self._simulation_id} is now running")

        with Timer() as timer:

            output: Dict[str, np.ndarray] = self._run_call()

        with open_component_file(self._out_file, self._simulation_id) as f:

            f.attrs['run_time'] = timer.total_time

            f.create_dataset(
                "parameters", data=np.array(list(self._parameter_set.values()))
            )

            for i in range(self._num_outputs):

                f.create_dataset(f"output_{i}", data=output[f"output_{i}"])

            if self._num_meta_parameters is not None:

                for i in range(self._num_meta_parameters):

                    f.attrs[f'meta_{i}'] = output[f"meta_{i}"]

    @abstractmethod
    def _run_call(self) -> Dict[str, np.ndarray]:

        log.error("Attempting to use base class")

        raise RuntimeError()


def gather(file_name: str, current_size: int = 0, clean: bool = True) -> None:

    # gather the list of files

    p = Path(file_name)

    if ronswanson_config.slurm.store_dir is None:

        parent_dir = p.absolute().parent

    else:

        parent_dir = Path(ronswanson_config.slurm.store_dir).absolute()

    multi_file_dir: Path = parent_dir / Path(f"{p.stem}_store")

    files = list(multi_file_dir.glob("sim_store_*.h5"))

    with h5py.File(file_name, "a") as database_file:

        database_file.attrs['has_been_touched'] = True

        for store in tqdm(
            files, desc="Gathering simulations", colour=Colors.GREEN.value
        ):

            log.debug(f"reading: {store}")

            sim_id: int = int(
                re.match("^sim_store_(\d*).h5", str(store.name)).groups()[0]
            )

            with h5py.File(str(store), "r") as f:

                database_file["parameters"][current_size + sim_id] = f[
                    "parameters"
                ][()]

                for k, v in database_file["values"].items():

                    v[current_size + sim_id] = f[k][()]

                database_file["run_time"][current_size + sim_id] = f.attrs[
                    "run_time"
                ]

                if "meta" in (database_file.keys()):

                    for k, v in database_file["meta"].items():

                        v[current_size + sim_id] = f.attrs[k]

            if clean:

                log.debug(f"removing: {store}")

                store.unlink()

    if clean:

        multi_file_dir.rmdir()



import numba as nb
import numpy as np


@nb.njit(cache=True)
def cartesian_jit(arrays):

    n = 1
    for x in arrays:
        n *= x.size
    out = np.zeros((n, len(arrays)))

    for i in range(len(arrays)):
        m = int(n / arrays[i].size)
        out[:n, i] = np.repeat(arrays[i], m)
        n //= arrays[i].size

    n = arrays[-1].size
    for k in range(len(arrays) - 2, -1, -1):
        n *= arrays[k].size
        m = int(n / arrays[k].size)
        for j in range(1, arrays[k].size):
            out[j * m : (j + 1) * m, k + 1 :] = out[0:m, k + 1 :]
    return out



from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, List, Optional
from pathlib import Path


import h5py
import numpy as np
import plotly.graph_objects as go
from astromodels import TemplateModel, TemplateModelFactory
from astromodels.utils.logging import silence_console_log
from tqdm.auto import tqdm

from ronswanson.grids import Parameter, ParameterGrid
from ronswanson.utils.color import Colors

from .utils.logging import setup_logger

from joblib import Parallel, delayed

log = setup_logger(__name__)


@dataclass(frozen=True)
class ValueContainer:
    params: np.ndarray
    values: np.ndarray


class Database:
    def __init__(
        self,
        grid_points: np.ndarray,
        parameter_names: List[str],
        energy_grid: np.ndarray,
        run_time: np.ndarray,
        values: np.ndarray,
        meta_data: Optional[Dict[str, np.ndarray]] = None,
    ) -> None:
        """
        Databse of parameters and simulated values

        :param grid_points:
        :type grid_points: Dict[str, np.ndarray]
        :param parameter_names:
        :type parameter_names: List[str]
        :param energy_grid:
        :type energy_grid: np.ndarray
        :param values:
        :type values: Dict[str, np.ndarray]
        :returns:

        """

        self._n_entries = len(values)

        self._n_parameters: int = len(parameter_names)

        self._parameter_names: List[str] = parameter_names

        self._energy_grid: np.ndarray = energy_grid

        # self._values: Dict[str, np.ndarray] = values
        # self._grid_points: Dict[str, np.ndarray] = grid_points

        self._values: np.ndarray = values

        self._grid_points: np.ndarray = grid_points

        self._run_time: np.ndarray = run_time

        self._parameter_ranges: Dict[str, np.ndarray] = OrderedDict()

        for i, par in enumerate(self._parameter_names):

            self._parameter_ranges[par] = np.sort(
                np.unique(self._grid_points[:, i])
            )

        self._meta_data: Optional[Dict[str, np.ndarray]] = meta_data

        if np.any(~np.isfinite(self._values)):

            log.error("The table values contain non-finite values")
            log.info(
                "you can replace these by calling the replace_nan_inf_with() member"
            )

    @property
    def n_entries(self) -> int:
        """
        the number of entries in the database

        :returns:

        """

        return self._n_entries

    @property
    def n_parameters(self) -> int:
        """
        the number of parameters in the database

        :returns:

        """

        return self._n_parameters

    @property
    def grid_points(self) -> np.ndarray:
        return self._grid_points

    @property
    def values(self) -> np.ndarray:
        return self._values

    @property
    def parameter_ranges(self) -> Dict[str, np.ndarray]:
        return self._parameter_ranges

    @property
    def parameter_names(self) -> List[str]:
        """
        the names of the parameters
        """
        return self._parameter_names

    @property
    def run_time(self) -> np.ndarray:
        """
        an array of runtimes for all points in the grid

        :returns:

        """

        return self._run_time

    @property
    def energy_grid(self) -> np.ndarray:
        """
        The corresponding energy grid of the simulation

        :returns:

        """

        return self._energy_grid

    @property
    def meta_data(self) -> Optional[Dict[str, np.ndarray]]:

        return self._meta_data

    def at(self, i: int) -> ValueContainer:
        """
        return the parameters and values
        at an index
        """

        return ValueContainer(
            params=self._grid_points[i, :], values=self._values[i, :]
        )

    def _sub_selection(
        self,
        paramater_name: str,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> np.ndarray:

        if paramater_name not in self._parameter_names:

            log.error(f"{paramater_name} is not in {self._parameter_names}")

            raise AssertionError()

        par_idx: int = self._parameter_names.index(paramater_name)

        selection = np.ones(self._n_entries, dtype=bool)

        if vmin is not None:

            selection = (self._grid_points[..., par_idx] >= vmin) & selection

        if vmax is not None:

            selection = (self._grid_points[..., par_idx] <= vmax) & selection

        return selection

    def _parameter_sub_selection(
        self,
        paramater_name: str,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> np.ndarray:

        if paramater_name not in self._parameter_names:

            log.error(f"{paramater_name} is not in {self._parameter_names}")

            raise AssertionError()

        par_range = self._parameter_ranges[paramater_name]

        selection = np.ones(len(par_range), dtype=bool)

        if vmin is not None:

            selection = (par_range >= vmin) & selection

        if vmax is not None:

            selection = (par_range <= vmax) & selection

        return selection

    @classmethod
    def from_file(cls, file_name: str, output: int = 0) -> "Database":
        """
        open a database from a file.
        The output argument specifies which value to collect.
        For example, a simulation may save photon and electron
        distributions from a solution.
        """
        values = {}
        parameters = {}

        with h5py.File(file_name, "r") as f:

            energy_grid = f['energy_grid'][f'energy_grid_{output}'][()]

            values_grp = f["values"]

            run_time = f["run_time"][()]

            par_name_grp = f["parameter_names"]

            parameter_names = [
                par_name_grp.attrs[f"par{i}"]
                for i in range(len(par_name_grp.attrs))
            ]

            parameters = f['parameters'][()]

            values = values_grp[f'output_{output}'][()]

            meta_data = None

            if "meta" in list(f.keys()):

                meta_data = {}

                for k, v in f['meta'].items():

                    meta_data[k] = v[()]

        return cls(
            grid_points=parameters,
            parameter_names=parameter_names,
            energy_grid=energy_grid,
            values=values,
            run_time=run_time,
            meta_data=meta_data,
        )

    def replace_nan_inf_with(self, value: float = 0.0) -> None:

        """
        Replace NaN and inf values with a float

        :param value:
        :type value: float
        :returns:

        """
        idx = np.isinf(self._values) | np.isnan(self._values)

        self._values[idx] = value

    def to_3ml(
        self,
        name: str,
        desc: str,
        overwrite: bool = False,
        **kwargs,
    ) -> TemplateModel:

        """
        construct a table model from the database.
        parameter sub-selections are passed as kwargs of
        dictionaries:

        selections = dict(param1=dict(vmin=1, vmax=2))

        :param name:
        :type name: str
        :param desc:
        :type desc: str
        :param overwrite:
        :type overwrite: bool
        :returns:

        """
        selection = np.ones(self._n_entries, dtype=bool)

        parameter_selection = {}

        for k, v in self._parameter_ranges.items():

            parameter_selection[k] = np.ones(len(v), dtype=bool)

        for k, v in kwargs.items():

            if k in self._parameter_names:

                vmin = None
                vmax = None

                if 'vmin' in v:

                    vmin = v['vmin']

                if 'vmax' in v:

                    vmax = v['vmax']

                selection = selection & self._sub_selection(k, vmin, vmax)

                parameter_selection[k] = parameter_selection[
                    k
                ] & self._parameter_sub_selection(k, vmin, vmax)

        # sub selections if any

        sub_grid = self._grid_points[selection, ...]
        sub_values = self._values[selection, ...]

        sub_parameter_ranges = {}

        for k, v in parameter_selection.items():

            sub_parameter_ranges[k] = self._parameter_ranges[k][
                parameter_selection[k]
            ]

        tmf = TemplateModelFactory(
            name, desc, self._energy_grid, self._parameter_names
        )

        for k, v in sub_parameter_ranges.items():

            tmf.define_parameter_grid(k, sub_parameter_ranges[k])

        with silence_console_log():

            for i in tqdm(
                range(len(sub_values)),
                desc="building table model",
                colour=Colors.YELLOW.value,
            ):

                ### DO NOT SORT

                tmf.add_interpolation_data(
                    sub_values[i],
                    **{
                        k: v for k, v in zip(self._parameter_names, sub_grid[i])
                    },
                )

            tmf.save_data(overwrite=overwrite)

        return TemplateModel(name)

    def check_for_missing_parameters(
        self, parameter_grid: ParameterGrid, create_new_grid: bool = False
    ) -> None:

        """Search a parameter grid for missing values

        :param parameter_grid:
        :type parameter_grid: ParameterGrid
        :param create_new_grid:
        :type create_new_grid: bool
        :returns:

        """
        missing_parameters = []

        for i in tqdm(
            range(parameter_grid.n_points),
            desc="search through parameter grid",
            colour=Colors.BLUE.value,
        ):

            these_parameters = np.atleast_2d(
                parameter_grid.at_index(i, as_array=True)
            )

            # see if these parmeters exist!

            if (
                np.isclose(self._grid_points, these_parameters).all(-1).sum()
                == 0
            ):

                log.debug(f"MISSING:{these_parameters[0]}")

                missing_parameters.append(these_parameters[0].tolist())

        if len(missing_parameters) == 0:

            log.info("There are no missing parameters!")

        else:

            log.warning(
                f"There were {len(missing_parameters)} missing parameters!"
            )

            if create_new_grid:

                missing_parameters = np.array(missing_parameters)

                log.info("Creating a new grid")

                parameter_list = []

                for i, par_name in enumerate(parameter_grid.parameter_names):

                    par = Parameter(
                        name=par_name,
                        custom=True,
                        values=missing_parameters[:, i],
                    )

                    parameter_list.append(par)

                new_parameter_grid = ParameterGrid(
                    parameter_list, parameter_grid.energy_grid
                )

                new_parameter_grid.write("missing_parameter_grid.yml")

    def _parallel_coord_plot(
        self,
        parameter: np.ndarray,
        colorscale: str = "viridis",
        as_log: bool = False,
    ) -> None:

        dims = [
            dict(label=p, values=v)
            for p, v in zip(self._parameter_names, self._grid_points.T)
        ]

        if as_log:

            parameter = np.log10(parameter)

        fig = go.Figure(
            data=go.Parcoords(
                line=dict(
                    color=parameter,
                    colorscale=colorscale,
                    showscale=True,
                    cmin=parameter.min(),
                    cmax=parameter.max(),
                ),
                dimensions=dims,
                unselected=dict(line=dict(color='white', opacity=0.01)),
            )
        )

        fig.show()

    def plot_runtime(self, colorscale: str = "plasma"):

        """
        show a parallel plot of the run time

        :param colorscale:
        :type colorscale:
        :returns:

        """
        self._parallel_coord_plot(
            self._run_time, colorscale=colorscale, as_log=True
        )

    def plot_meta_data(
        self,
        meta_number: int = 0,
        colorscale: str = "plasma",
        as_log: bool = False,
    ):

        """

        parallel plot of the meta data

        :param meta_number:
        :type meta_number: int
        :param colorscale:
        :type colorscale: str
        :param as_log:
        :type as_log: bool
        :returns:

        """

        if self._meta_data is not None:

            self._parallel_coord_plot(
                self._meta_data[f"meta_{meta_number}"],
                colorscale=colorscale,
                as_log=as_log,
            )

        else:

            msg = "This database has no meta data"

            log.error(msg)

            raise RuntimeError(msg)


def update_database(
    database_filename: str,
    sim_number_to_replace: np.ndarray,
    sim_locations: str,
    create_backup: bool = True,
) -> None:

    """
    will replace the simulations (value arrays) of the indices supplied
    from the folder indicated. This is useful when certain runs of a
    simulation failed and one wants to replace those.

    Note: this requires that the sim ids of the replacement runs match those
    in the database!

    :param database_filename:
    :type database_filename: str
    :param sim_number_to_replace:
    :type sim_number_to_replace: np.ndarray
    :param sim_locations:
    :type sim_locations: str
    :returns:

    """
    import shutil

    database_path: Path = Path(database_filename).absolute()

    if create_backup:

        backup_file_name = (
            database_path.parents / f"{database_path.stem}_bkup.h5"
        )

        log.info(f"creating a backup as {backup_file_name}")

        shutil.copy(database_path, backup_file_name)

    with open(database_path.as_posix(), "r+") as f:

        n_output = 0
        for key in list(f.keys()):

            if "output_" in key:
                n_output += 1

        n_meta = 0
        for key in list(f.attrs.keys()):
            if "meta_" in key:
                n_meta += 1

        for idx in tqdm(sim_number_to_replace, desc="replaceing values"):

            with open(
                Path(sim_locations).absolute() / f"sim_store_{idx}.h5", "r"
            ) as r:

                for i in range(n_output):

                    f[f"values/output_{i}"][idx, :] = r[f"output_{i}"][()]

                for i in range(n_meta):

                    f[f"meta/meta_{i}"][idx, :] = r.attrs[f"meta_{i}"]

                f["runtime"][idx] = r.attrs["runtime"]


def merge_databases(
    *file_names: List[str], new_name: str = "merged_db.h5"
) -> None:

    """TODO describe function

    :param new_name:
    :type new_name: str
    :returns:

    """
    n_entries = 0

    for i, fname in enumerate(file_names):

        with h5py.File(fname, "r") as f:

            n_entries += f["parameters"].shape[0]

            if i == 0:

                par_name_grp = f["parameter_names"]

                parameter_names = [
                    par_name_grp.attrs[f"par{i}"]
                    for i in range(len(par_name_grp.attrs))
                ]

                energy_grids = []

                for _, v in f["energy_grid"].items():

                    energy_grids.append(v[()])

        n_outputs = len(energy_grids)

        n_parameters = len(parameter_names)

        parameters = np.zeros((n_entries, n_parameters))

        values = []

        for grid in energy_grids:

            values.append(np.zeros((n_entries, len(grid))))

        k = 0

        # extract all the information

        for i, fname in enumerate(file_names):

            with h5py.File(fname, "r") as f:

                n = f["parameters"].shape[0]

                parameter_names[k:n, :] = f["parameters"][()]

                for j, val in enumerate(values):

                    val[k:n] = f["values"][f"output_{j}"][()]

                k += n

        # create the new file

        with h5py.File(new_name, "w") as f:

            f.create_dataset("parameters", data=parameters)

            val_grp = f.create_group("values")
            energy_grid_grp = f.create_group("energy_grid")

            par_name_grp = f.create_group("parameters_names")
            for i in range(n_parameters):

                par_name_grp.attrs[f"par{i}"] = parameter_names[i]

            for i in range(n_outputs):

                val_grp.create_dataset(f"output_{i}", data=values[i])
                energy_grid_grp.create_dataset(
                    f"energy_grid_{i}", data=energy_grids[i]
                )


__all__ = ["Database"]



import matplotlib.colors as mcolors
import numpy as np


def hex_to_rgb(value):
    '''
    Converts hex to rgb colours
    value: string of 6 characters representing a hex colour.
    Returns: list length 3 of RGB values'''
    value = value.strip("#")  # removes hash symbol if present
    lv = len(value)
    return tuple(int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3))


def rgb_to_dec(value):
    '''
    Converts rgb to decimal colours (i.e. divides each value by 256)
    value: list (length 3) of RGB values
    Returns: list (length 3) of decimal values'''
    return [v / 256 for v in value]


def get_continuous_cmap(hex_list, float_list=None):

    rgb_list = [rgb_to_dec(hex_to_rgb(i)) for i in hex_list]
    if float_list:
        pass
    else:
        float_list = list(np.linspace(0, 1, len(rgb_list)))

    cdict = dict()
    for num, col in enumerate(['red', 'green', 'blue']):
        col_list = [
            [float_list[i], rgb_list[i][num], rgb_list[i][num]]
            for i in range(len(float_list))
        ]
        cdict[col] = col_list
    cmp = mcolors.LinearSegmentedColormap('my_cmp', segmentdata=cdict, N=256)
    return cmp



import json
import shutil
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import h5py
import numpy as np
import numba as nb
import yaml
from omegaconf import MISSING, OmegaConf
from tqdm.auto import tqdm

from ronswanson.utils.color import Colors
from ronswanson.utils.check_complete import check_complete_ids

from .grids import ParameterGrid
from .script_generator import (
    PythonGatherGenerator,
    PythonGenerator,
    SLURMGatherGenerator,
    SLURMGenerator,
)
from .utils.logging import setup_logger

log = setup_logger(__name__)


@dataclass
class SLURMTime:
    hrs: int = 0
    min: int = 10
    sec: int = 0


@dataclass
class JobConfig:
    time: SLURMTime
    n_cores_per_node: int


@dataclass
class GatherConfig(JobConfig):
    n_gather_per_core: int


@dataclass
class SimulationConfig(JobConfig):
    n_mp_jobs: int
    run_per_node: Optional[int] = None
    use_nodes: bool = False
    max_nodes: Optional[int] = None
    linear_execution: bool = False
    num_meta_parameters: Optional[int] = None


## structure for file


@dataclass
class JobConfigStructure:
    time: Optional[SLURMTime] = None
    n_cores_per_node: Optional[int] = None


@dataclass
class GatherConfigStructure(JobConfigStructure):
    n_gather_per_core: int = MISSING


@dataclass
class SimulationConfigStructure(JobConfigStructure):
    n_mp_jobs: int = MISSING
    run_per_node: Optional[int] = None
    use_nodes: bool = False
    max_nodes: Optional[int] = None
    linear_execution: bool = False


@dataclass
class YAMLStructure:
    import_line: str = MISSING
    parameter_grid: str = MISSING
    out_file: str = MISSING
    clean: bool = True
    simulation: SimulationConfigStructure = SimulationConfigStructure()
    gather: Optional[GatherConfigStructure] = None
    num_meta_parameters: Optional[int] = None
    finish_missing: bool = False


class SimulationBuilder:
    """
    The simulation builder class constructs the scripts
    needed for building the table model database

    """

    def __init__(
        self,
        parameter_grid: ParameterGrid,
        out_file: str,
        import_line: str,
        simulation_config: SimulationConfig,
        gather_config: Optional[GatherConfig] = None,
        num_meta_parameters: Optional[int] = None,
        clean: bool = True,
        finish_missing: bool = False,
    ):

        """TODO describe function

        :param parameter_grid:
        :type parameter_grid: ParameterGrid
        :param out_file:
        :type out_file: str
        :param import_line:
        :type import_line: str
        :param simulation_config:
        :type simulation_config: SimulationConfig
        :param gather_config:
        :type gather_config: Optional[GatherConfig]
        :param num_meta_parameters:
        :type num_meta_parameters: Optional[int]
        :param clean:
        :type clean: bool
        :returns:

        """
        self._has_complete_params: bool = False

        self._import_line: str = import_line

        self._simulation_config: SimulationConfig = simulation_config

        self._gather_config: Optional[GatherConfig] = gather_config

        self._out_file: str = out_file

        self._base_dir: Path = Path(out_file).parent.absolute()

        self._num_meta_parameters: Optional[int] = num_meta_parameters

        # write out the parameter file

        self._parameter_file: Path = self._base_dir / "parameters.yml"

        parameter_grid.write(str(self._parameter_file))

        self._n_outputs: int = len(parameter_grid.energy_grid)

        self._clean: bool = clean

        self._n_iterations: int = parameter_grid.n_points

        self._current_database_size: int = 0

        self._finish_missing: bool = finish_missing

        if not self._finish_missing:

            self._initialize_database()

        # if we are using nodes
        # we need to see how many
        # we need

        log.info(
            f"there are [bold bright_red]{self._n_iterations} iterations [/bold bright_red]"
        )

        if self._simulation_config.use_nodes:

            self._compute_chunks()

        else:

            self._n_nodes: Optional[int] = None

            self._n_gather_nodes: Optional[int] = None

        self._generate_python_script()

        if self._simulation_config.use_nodes:

            self._generate_slurm_script()

            output_dir = self._base_dir / "output"

            if not output_dir.exists():

                output_dir.mkdir()

                log.debug("created the output directory")

    @classmethod
    def from_yaml(cls, file_name: str) -> "SimulationBuilder":
        """
        Create a simulation setup from a yaml file.

        """

        # check the file structure

        structure = OmegaConf.structured(YAMLStructure)

        try:

            test_structure = OmegaConf.load(file_name)

            merged = OmegaConf.merge(structure, test_structure)

            OmegaConf.to_container(merged, throw_on_missing=True)

        except Exception as e:

            log.error(e)

            raise e

        with Path(file_name).open("r") as f:

            inputs = yaml.load(stream=f, Loader=yaml.SafeLoader)

        log.debug("reading setup inputs:")

        for k, v in inputs.items():

            log.debug(f"{k}: {v}")

        parameter_grid = ParameterGrid.from_yaml(inputs.pop("parameter_grid"))

        simulation_input = inputs.pop("simulation")

        if "time" in simulation_input:

            sim_time = SLURMTime(**simulation_input.pop("time"))

        else:

            sim_time = SLURMTime()

        if "n_cores_per_node" not in simulation_input:

            if "use_nodes" in simulation_input:

                if simulation_input["use_nodes"]:

                    log.warning(
                        "you are using nodes but did not specify the number of n_cores_per_node"
                    )
                    log.warning(
                        "the number of cores will be set to the number of multi-process jobs"
                    )

            simulation_input["n_cores_per_node"] = simulation_input["n_mp_jobs"]

        simulation_config = SimulationConfig(time=sim_time, **simulation_input)

        gather_config = None

        if "gather" in inputs:

            gather_inputs = inputs.pop("gather")

            gather_time = SLURMTime(**gather_inputs.pop("time"))

            gather_config = GatherConfig(time=gather_time, **gather_inputs)

        return cls(
            parameter_grid=parameter_grid,
            simulation_config=simulation_config,
            gather_config=gather_config,
            **inputs,
        )

    def _initialize_database(self) -> None:

        if not Path(self._out_file).exists():

            with h5py.File(self._out_file, "w") as f:

                f.attrs["has_been_touched"] = False

                pg = ParameterGrid.from_yaml(self._parameter_file)

                # store the parameter names

                p_name_group = f.create_group("parameter_names")

                for i, name in enumerate(pg.parameter_names):

                    p_name_group.attrs[f"par{i}"] = name

                # store the energy grids

                ene_grp = f.create_group("energy_grid")

                for i, grid in enumerate(pg.energy_grid):

                    ene_grp.create_dataset(
                        f"energy_grid_{i}",
                        data=grid.grid,
                        compression="gzip",
                    )

                # create an empty group for the parameters

                f.create_dataset(
                    "parameters",
                    shape=(pg.n_points,) + np.array(pg.parameter_names).shape,
                    maxshape=(None,) + np.array(pg.parameter_names).shape,
                    #    compression="gzip",
                )

                val_grp: h5py.Group = f.create_group("values")

                # create an empty data set for the values

                for i in range(len(pg.energy_grid)):

                    val_grp.create_dataset(
                        f"output_{i}",
                        shape=(pg.n_points,) + pg.energy_grid[i].grid.shape,
                        maxshape=(None,) + pg.energy_grid[i].grid.shape,
                        # compression="gzip",
                    )

                f.create_dataset(
                    "run_time", shape=(pg.n_points,), maxshape=(None,)
                )

                if self._num_meta_parameters is not None:

                    meta_grp: h5py.Group = f.create_group("meta")

                    log.debug("detected meta parameters")

                    for i in range(self._num_meta_parameters):

                        meta_grp.create_dataset(
                            f"meta_{i}", shape=(pg.n_points,), maxshape=(None,)
                        )

        else:

            # we need to resize the dataset

            log.warning(
                f"There was already a database: [red]{self._out_file}[/red]"
            )

            with h5py.File(self._out_file, "r") as f:

                has_been_touched = f.attrs["has_been_touched"]

            if not has_been_touched:

                log.warning("the database has not been gathered")
                log.warning("erasing and starting over")

                Path(self._out_file).unlink()

                time.sleep(2)

                self._initialize_database()

            copy_file_name: str = f"{Path(self._out_file).parent}/{Path(self._out_file).stem}_copy{Path(self._out_file).suffix}"

            log.warning(f"a copy will be made to [blue]{copy_file_name}[\blue]")

            shutil.copy(self._out_file, copy_file_name)

            self._check_completed()

            with h5py.File(self._out_file, "a") as f:

                pg = ParameterGrid.from_yaml(self._parameter_file)

                dataset: h5py.Dataset = f["parameters"]

                self._current_database_size = dataset.shape[0]

                log.warning(
                    f"The existing data base had {self._current_database_size} entries"
                )

                dataset.resize(
                    (self._current_database_size + pg.n_points,)
                    + dataset.shape[1:]
                )

                val_grp = f["values"]

                for i in range(len(pg.energy_grid)):

                    dataset: h5py.Dataset = val_grp[f"output_{i}"]

                    dataset.resize(
                        (self._current_database_size + pg.n_points,)
                        + dataset.shape[1:]
                    )

                dataset: h5py.Dataset = f["run_time"]

                dataset.resize((self._current_database_size + pg.n_points,))

                if self._num_meta_parameters is not None:

                    log.debug("detected meta parameters")

                    meta_grp = f["meta"]

                    for i in range(self._num_meta_parameters):

                        dataset: h5py.Dataset = meta_grp[f"meta_{i}"]

                        dataset.resize(
                            (self._current_database_size + pg.n_points,)
                        )

        self._n_outputs: int = len(pg.energy_grid)

    def _check_completed(self) -> None:

        if Path(self._out_file).exists():

            with h5py.File(self._out_file, "r") as f:

                params = f["parameters"][()]

            out_file = self._base_dir / "completed_parameters.json"

            with out_file.open("w") as f:

                json.dump(params.tolist(), f)

            self._has_complete_params = True

    def _compute_complete_ids(self):

        log.info("seeing how many are missing from the run")

        complete_ids = check_complete_ids(self._out_file)

        number_missing = self._n_iterations - len(complete_ids)

        log.info(f"there were {number_missing} runs")

        return np.array(complete_ids)

    def _compute_chunks(self) -> None:

        if self._finish_missing:

            complete_ids = self._compute_complete_ids()

            full = np.array(range(self._n_iterations))

            full[complete_ids] = -99

            incomplete_ids = full[full >= 0]

        else:

            complete_ids = []

        # we may only be cleaning up missing runs

        total_iterations: int = self._n_iterations - len(complete_ids)

        if self._simulation_config.run_per_node is None:

            log.debug("Each node will only execute the number of mp jobs")

            runs_per_node = 1

            generator = range(self._simulation_config.n_mp_jobs)

            n_nodes = np.ceil(
                total_iterations / self._simulation_config.n_mp_jobs
            )

        else:

            runs_per_node = self._simulation_config.run_per_node

            generator = range(runs_per_node)

            n_nodes = np.ceil(total_iterations / runs_per_node)

        if self._simulation_config.use_nodes:

            self._n_nodes = int(n_nodes)

            log.info(
                f"we will be using {self._n_nodes} nodes for the simulation"
            )

        # now generate the key files
        k = 0

        key_out = {}

        for i in tqdm(
            range(self._n_nodes),
            desc="computing node layout",
            colour=Colors.RED.value,
        ):

            output = []

            for j in generator:

                if not self._finish_missing:

                    if k < self._n_iterations:

                        output.append(k)

                else:

                    if k < total_iterations:

                        output.append(int(incomplete_ids[k]))

                k += 1

            key_out[i] = output

        with open(self._base_dir / "key_file.json", "w") as f:

            json.dump(key_out, f)

        # now collect the gather information

        if self._simulation_config.use_nodes:

            self._n_gather_nodes = int(
                np.ceil(
                    self._n_iterations
                    / (
                        self._gather_config.n_cores_per_node
                        * self._gather_config.n_gather_per_core
                    )
                )
            )

            rank_list = {}
            n = 0

            log.info(f"the gather task will use: {self._n_gather_nodes} nodes")
            log.debug(
                f"total_ranks: {self._n_gather_nodes * self._gather_config.n_cores_per_node}"
            )
            log.debug(f"number iterations: {self._n_iterations}")

            for i in tqdm(
                range(
                    self._n_gather_nodes * self._gather_config.n_cores_per_node
                ),
                desc="computing nodes for gather operation",
                colour=Colors.GREEN.value,
            ):

                core_list = []

                for j in range(self._gather_config.n_gather_per_core):

                    if n < self._n_iterations:

                        core_list.append(n)

                        n += 1

                rank_list[i] = core_list

            if not self._finish_missing:

                with open(self._base_dir / "gather_file.json", "w") as f:

                    json.dump(rank_list, f)

    def _generate_python_script(self) -> None:

        py_gen: PythonGenerator = PythonGenerator(
            "run_simulation.py",
            self._out_file,
            str(self._parameter_file),
            self._base_dir,
            self._import_line,
            self._simulation_config.n_mp_jobs,
            self._n_nodes,
            self._simulation_config.linear_execution,
            self._has_complete_params,
            self._current_database_size,
            clean=self._clean,
        )

        py_gen.write(str(self._base_dir))

        log.info(
            "[bold green blink]generated:[/bold green blink] run_simulation.py"
        )

    def _generate_slurm_script(self) -> None:

        multi_script: bool = False

        if self._simulation_config.max_nodes is not None:

            if self._n_nodes > self._simulation_config.max_nodes:

                log.debug("The number of reuested nodes is too large.")

                multi_script = True

                n_files = int(
                    np.ceil(self._n_nodes / self._simulation_config.max_nodes)
                )

                start = []
                stop = []
                current_number = 0

                for i in range(n_files):

                    start.append(current_number)

                    next_number = int(
                        (i + 1) * self._simulation_config.max_nodes
                    )

                    if next_number <= self._n_nodes:

                        stop.append(next_number)

                        current_number = next_number

                    else:

                        stop.append(self._n_nodes)

                        break

        if multi_script:

            for i, (a, b) in enumerate(zip(start, stop)):

                file_name = f"run_simulation_{i}.sh"

                slurm_gen: SLURMGenerator = SLURMGenerator(
                    file_name,
                    self._simulation_config.n_mp_jobs,
                    self._simulation_config.n_cores_per_node,
                    b,
                    self._simulation_config.time.hrs,
                    self._simulation_config.time.min,
                    self._simulation_config.time.sec,
                    node_start=a,
                )

                slurm_gen.write(str(self._base_dir))

                log.info(
                    f"[bold green blink]generated:[/bold green blink] {file_name}"
                )

        else:

            slurm_gen: SLURMGenerator = SLURMGenerator(
                "run_simulation.sh",
                self._simulation_config.n_mp_jobs,
                self._simulation_config.n_cores_per_node,
                self._n_nodes,
                self._simulation_config.time.hrs,
                self._simulation_config.time.min,
                self._simulation_config.time.sec,
            )

            slurm_gen.write(str(self._base_dir))

            log.info(
                "[bold green blink]generated:[/bold green blink] run_simulations.sh"
            )

        if not self._finish_missing:

            slurm_gen: SLURMGatherGenerator = SLURMGatherGenerator(
                "gather_results.sh",
                self._gather_config.n_cores_per_node,
                self._n_gather_nodes,
                self._gather_config.time.hrs,
                self._gather_config.time.min,
                self._gather_config.time.sec,
            )

            slurm_gen.write(str(self._base_dir))

            log.info(
                "[bold green blink]generated:[/bold green blink] gather_results.sh"
            )

            python_gather_gen: PythonGatherGenerator = PythonGatherGenerator(
                "gather_results.py",
                database_file_name=self._out_file,
                current_size=self._current_database_size,
                n_outputs=self._n_outputs,
                clean=self._clean,
                num_meta_parameters=self._num_meta_parameters,
            )

            python_gather_gen.write(str(self._base_dir))

            log.info(
                "[bold green blink]generated:[/bold green blink] gather_results.py"
            )



import itertools
import re
from pathlib import Path
from typing import List, Dict, Optional

import matplotlib.colors as mcolors
import matplotlib
import numpy as np
import plotille
from tqdm.auto import tqdm

from ronswanson.grids import ParameterGrid
from ronswanson.utils.color import Colors
from ronswanson.utils.colormap_generator import get_continuous_cmap
from ronswanson.utils.logging import setup_logger

from .configuration import ronswanson_config

log = setup_logger(__name__)


def check_complete_ids(database_file_name: str) -> List[int]:

    p = Path(database_file_name)

    if ronswanson_config.slurm.store_dir is None:

        parent_dir = p.absolute().parent

    else:

        parent_dir = Path(ronswanson_config.slurm.store_dir).absolute()

    multi_file_dir: Path = parent_dir / Path(f"{p.stem}_store")

    finished_ids: List[int] = []

    for fname in tqdm(
        list(multi_file_dir.glob("sim_store_*.h5")),
        desc="checking existing files",
        colour=Colors.GREEN.value,
    ):

        sim_id = int(re.search("sim_store_(\d*).h5", str(fname)).groups()[0])

        finished_ids.append(sim_id)

    finished_ids = np.array(finished_ids)

    return finished_ids


def int_formatter(val, chars, delta, left):
    return '{:{}{}}'.format(int(val), '<' if left else '>', chars)


def make_fig(
    this_name: str,
    other_name: str,
    finished_values: np.ndarray,
    unfinished_values: np.ndarray,
    width: int,
) -> str:

    fig = plotille.Figure()

    fig.register_label_formatter(float, int_formatter)
    fig.color_mode = 'rgb'
    fig.width = int(np.ceil(width))
    fig.height = 10
    fig.x_label = this_name
    fig.y_label = other_name

    fig.scatter(
        unfinished_values[:, 0],
        unfinished_values[:, 1],
        lc="FE093A",
        marker="o",
    )

    fig.scatter(
        finished_values[:, 0],
        finished_values[:, 1],
        lc="2DFF96",
        marker="o",
    )

    print(fig.show())


def examine_parameter(
    database_file_name: str,
    parameter_grid_file_name: str,
    parameter_to_check: str,
) -> None:

    finished_ids = check_complete_ids(database_file_name)

    pg = ParameterGrid.from_yaml(parameter_grid_file_name)

    if parameter_to_check not in pg.parameter_names:

        msg = f"{parameter_to_check} is not a valid parameter name"

        log.error(msg)

        raise RuntimeError(msg)

    idx = pg.parameter_names.index(parameter_to_check)

    lines: List[str] = []

    fig_width = int(np.ceil(100.0 / (pg.n_parameters - 1)))
    fig_width = 40

    n_points = pg.n_points

    for jdx, other_name in enumerate(pg.parameter_names):

        # do not check against yourself
        if idx != jdx:

            finished_values = []
            unfinished_values = []

            k = 0

            for i in range(n_points):

                result = pg.full_grid[i]

                this_param = result[idx]
                other_param = result[jdx]

                if k in finished_ids:

                    finished_values.append([this_param, other_param])

                else:

                    unfinished_values.append([this_param, other_param])

                k += 1

            finished_values = np.array(finished_values)
            unfinished_values = np.array(unfinished_values)

            line = make_fig(
                parameter_to_check,
                other_name=other_name,
                finished_values=finished_values,
                unfinished_values=unfinished_values,
                width=fig_width,
            )

            lines.append(line)


def make_fig_detailed(
    this_name: str,
    other_name: str,
    this_param_values: np.ndarray,
    other_param_values: np.ndarray,
    count_dict: Dict[str, Dict[str, int]],
    n_points: int,
    width: int,
    colormap,
) -> None:

    fig = plotille.Figure()

    fig.register_label_formatter(float, int_formatter)
    fig.color_mode = 'rgb'
    fig.width = int(np.ceil(width))
    fig.height = 10
    fig.x_label = this_name
    fig.y_label = other_name

    total_other = n_points / (len(this_param_values) * len(other_name))

    for result in itertools.product(this_param_values, other_param_values):

        a, b = result

        count = 0

        if a in count_dict:

            if b in count_dict[a]:

                count = count_dict[a][b]

        color = mcolors.to_hex(colormap(float(count / total_other)))[1:]

        fig.scatter(
            [a],
            [b],
            lc=color,
            marker="o",
        )

    print(fig.show())


def examine_parameter_detailed(
    database_file_name: str,
    parameter_grid_file_name: str,
    parameter_to_check: str,
    colormap: Optional[str] = None,
) -> None:

    if colormap is None:

        cmap = get_continuous_cmap(['#FE093A', '#FFFF53', '#2DFF96'])

    else:

        cmap = matplotlib.cm.get_cmap(colormap)

    finished_ids = check_complete_ids(database_file_name)

    pg = ParameterGrid.from_yaml(parameter_grid_file_name)

    if parameter_to_check not in pg.parameter_names:

        msg = f"{parameter_to_check} is not a valid parameter name"

        log.error(msg)

        raise RuntimeError(msg)

    idx = pg.parameter_names.index(parameter_to_check)

    lines: List[str] = []

    fig_width = int(np.ceil(100.0 / (pg.n_parameters - 1)))
    fig_width = 40

    n_points = pg.n_points

    for jdx, other_name in enumerate(pg.parameter_names):

        counts = np.zeros((len(finished_ids), 2))

        # do not check against yourself
        if idx != jdx:

            finished_values = []
            unfinished_values = []

            for i, j in enumerate(finished_ids):

                result = pg.full_grid[j]

                this_param = result[idx]
                other_param = result[jdx]

                counts[i] = np.array([this_param, other_param])

            unique, counts = np.unique(counts, axis=0, return_counts=True)

            count_dict = {}

            for i, u in enumerate(unique):

                if u[0] in count_dict:

                    count_dict[u[0]][u[1]] = counts[i]

                else:

                    inner_dict = {}
                    inner_dict[u[1]] = counts[i]
                    count_dict[u[0]] = inner_dict

            finished_values = np.array(finished_values)
            unfinished_values = np.array(unfinished_values)

            make_fig_detailed(
                parameter_to_check,
                other_name=other_name,
                this_param_values=pg.parameter_list[idx].grid,
                other_param_values=pg.parameter_list[jdx].grid,
                count_dict=count_dict,
                n_points=n_points,
                width=fig_width,
                colormap=cmap,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

import versioneer
from setuptools import setup


setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    entry_points={
        'console_scripts': [
            'simulation_build = scripts.simulation_build:simulation_build',
            'examine_simulation = scripts.check:examine_simulation',
            'examine_simulation_detailed = scripts.check:examine_simulation_detailed'
        ],
    },
    #        package_data={"": extra_files},
)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from ghost_writer import ScriptGenerator

from ronswanson.utils.logging import setup_logger

from .utils import ronswanson_config

log = setup_logger(__name__)


class PythonGenerator(ScriptGenerator):
    def __init__(
        self,
        file_name: str,
        database_file: str,
        parameter_file: str,
        base_dir: str,
        import_line: str,
        n_procs: int,
        n_nodes: Optional[int] = None,
        linear_exceution: bool = False,
        has_complete_params: bool = False,
        current_size: int = 0,
        clean: bool = True,
    ) -> None:

        """
        Generate the python script that will be run


        :param file_name:
        :type file_name: str
        :param database_file:
        :type database_file: str
        :param parameter_file:
        :type parameter_file: str
        :param base_dir:
        :type base_dir: str
        :param import_line:
        :type import_line: str
        :param n_procs:
        :type n_procs: int
        :param n_nodes:
        :type n_nodes: Optional[int]
        :param linear_exceution:
        :type linear_exceution: bool
        :returns:

        """
        self._import_line = import_line
        self._n_procs: int = n_procs
        self._n_nodes: Optional[int] = n_nodes
        self._parameter_file: str = parameter_file
        self._database_file: str = Path(database_file).absolute()
        self._base_dir: str = base_dir
        self._linear_execution: bool = linear_exceution
        self._has_complete_params: bool = has_complete_params
        self._current_size: int = current_size
        self._clean: bool = clean

        super().__init__(file_name)

    def _build_script(self) -> None:

        self._add_line(self._import_line)
        self._add_line("from joblib import Parallel, delayed")
        self._add_line("import json")
        self._add_line("import numpy as np")
        self._add_line("from tqdm.auto import tqdm")
        self._add_line("from ronswanson import ParameterGrid")
        self._add_line("from ronswanson.utils.logging import setup_logger")
        self._add_line("from ronswanson.utils import Colors")
        self._add_line("from ronswanson.simulation import gather")

        if self._n_nodes is not None:
            self._add_line("import sys")
            self._end_line()
            self._add_line("key_num = int(sys.argv[-1])")

        self._add_line("log = setup_logger(__name__)")

        self._end_line()

        if self._has_complete_params:

            self._add_line("with open('completed_parameters.json', 'r') as f:")
            self._add_line("complete_params = json.load(f)", indent_level=1)

        self._add_line(
            f"pg = ParameterGrid.from_yaml('{self._parameter_file}')"
        )

        self._add_line("def func(i):")
        self._add_line("params = pg.at_index(i)", indent_level=1)

        if self._has_complete_params:

            self._add_line("for p in complete_params:", indent_level=1)
            self._add_line(
                "if np.alltrue(np.array(p) == params):", indent_level=2
            )
            self._add_line(
                "log.debug('parameters already exists in file!')",
                indent_level=3,
            )
            self._add_line("return", indent_level=3)

        self._add_line(
            f"simulation = Simulation(i, params, pg.energy_grid,'{self._database_file}')",
            indent_level=1,
        )
        self._add_line("simulation.run()", indent_level=1)

        if self._n_nodes is None:

            self._add_line("iteration = [i for i in range(0, pg.n_points)]")

        else:

            self._add_line(
                f"with open(f'{self._base_dir}/key_file.json','r') as f:"
            )

            self._add_line("keys = json.load(f)[str(key_num)]", indent_level=1)

            self._add_line("iteration = [int(x) for x in keys]", indent_level=1)

            pass

        if self._linear_execution:

            # just do a straight for loop

            self._add_line("for i in tqdm(iteration):")
            self._add_line("func(i)", indent_level=1)

            self._add_line(
                f"gather('{self._database_file}', {self._current_size}, clean=True)"
            )

        else:

            # use joblib

            if self._n_nodes is not None:

                self._add_line(
                    f"Parallel(n_jobs={self._n_procs})(delayed(func)(i) for i in iteration)"
                )

            else:

                self._add_line(
                    f"Parallel(n_jobs={self._n_procs})(delayed(func)(i) for i in tqdm(iteration, colour=Colors.RED.value, desc='simulating function'))"
                )

                self._add_line(
                    f"gather('{self._database_file}', {self._current_size}, clean={self._clean})"
                )


class PythonGatherGenerator(ScriptGenerator):
    def __init__(
        self,
        file_name: str,
        database_file_name: str,
        current_size: int,
        n_outputs: int,
        num_meta_parameters: Optional[int] = None,
        clean: bool = True,
    ) -> None:

        self._database_file_name: str = database_file_name
        self._current_size: int = current_size
        self._n_outputs: int = n_outputs
        self._clean: bool = clean
        self._num_meta_parmeters: Optional[int] = num_meta_parameters

        super().__init__(file_name)

    def _build_script(self) -> None:
        self._add_line('import json')
        self._add_line('from mpi4py import MPI')
        self._end_line()
        self._add_line(
            'from ronswanson.utils.configuration import ronswanson_config'
        )
        self._end_line()
        self._add_line('import h5py')
        self._add_line('import sys')
        self._add_line('from pathlib import Path')
        self._end_line()
        self._end_line()
        self._add_line('rank = MPI.COMM_WORLD.rank')
        self._end_line()
        self._add_line('with open("gather_file.json", "r") as f:')
        self._end_line()
        self._add_line(
            'sim_ids = [int(x) for x in json.load(f)[str(rank)]]',
            indent_level=1,
        )
        self._end_line()
        self._end_line()
        self._add_line(f'p = Path("{self._database_file_name}").absolute()')
        self._end_line()
        self._end_line()
        self._add_line(
            'database = h5py.File(str(p), "a", driver="mpio", comm=MPI.COMM_WORLD)'
        )

        self._add_line('database.attrs["has_been_touched"] = True')
        self._end_line()
        self._end_line()
        self._end_line()
        self._add_line('if ronswanson_config.slurm.store_dir is None:')
        self._end_line()
        self._add_line('parent_dir = p.absolute().parent', indent_level=1)
        self._end_line()
        self._add_line('else:')
        self._end_line()
        self._add_line(
            'parent_dir = Path(ronswanson_config.slurm.store_dir).absolute()',
            indent_level=1,
        )
        self._end_line()
        self._add_line(
            'multi_file_dir: Path = parent_dir / Path(f"{p.stem}_store")'
        )
        self._end_line()
        self._end_line()
        self._add_line(f'current_size = {self._current_size}')
        self._end_line()
        self._end_line()
        self._add_line('db_params = database["parameters"]')
        self._add_line('db_run_time = database["run_time"]')
        self._end_line()
        self._add_line('vals = database["values"]')

        if self._num_meta_parmeters is not None:

            self._add_line("meta = database['meta']")

            for i in range(self._num_meta_parmeters):

                self._add_line(f'meta_{i} = meta["meta_{i}"]')

        self._end_line()

        for i in range(self._n_outputs):

            self._add_line(f'output_{i} = vals["output_{i}"]')
        self._end_line()

        self._end_line()
        self._end_line()
        self._add_line('for sim_id in sim_ids:')
        self._end_line()
        self._add_line(
            'this_file: Path = multi_file_dir / f"sim_store_{sim_id}.h5"',
            indent_level=1,
        )
        self._end_line()
        self._add_line('index = int(current_size + sim_id)', indent_level=1)
        self._end_line()
        self._add_line('with h5py.File(this_file, "r") as f:', indent_level=1)
        self._end_line()
        self._add_line(
            'db_params[index, :] = f["parameters"][()]', indent_level=2
        )
        self._end_line()

        self._add_line(
            'db_run_time[index] = f.attrs["run_time"]', indent_level=2
        )

        for i in range(self._n_outputs):

            self._add_line(
                f'output_{i}[index, :] = f["output_{i}"][()]', indent_level=2
            )

        if self._num_meta_parmeters is not None:

            for i in range(self._num_meta_parmeters):

                self._add_line(
                    f'meta_{i}[index] = f.attrs["meta_{i}"]', indent_level=2
                )

        if self._clean:

            self._add_line("this_file.unlink()", indent_level=1)

        self._end_line()
        self._end_line()
        self._end_line()
        self._add_line('f.close()')


class SLURMGenerator(ScriptGenerator):
    def __init__(
        self,
        file_name: str,
        n_procs: int,
        n_procs_to_use: int,
        n_nodes: int,
        hrs: int,
        min: int,
        sec: int,
        node_start: int = 0,
    ) -> None:

        self._n_procs: int = n_procs
        self._n_procs_to_use: int = n_procs_to_use
        self._n_nodes: int = n_nodes
        self._node_start: int = node_start
        self._hrs: int = hrs
        self._min: int = min
        self._sec: int = sec

        super().__init__(file_name)

    def _build_script(self) -> None:

        self._add_line("#!/bin/bash")
        self._add_line("")

        if self._node_start < (self._n_nodes - 1):

            self._add_line(
                f"#SBATCH --array={self._node_start}-{self._n_nodes-1} #generate array"
            )

        self._add_line("#SBATCH -o ./output/%A_%a.out      #output file")
        self._add_line("#SBATCH -e ./output/%A_%a.err      #error file")
        self._add_line("#SBATCH -D ./                      #working directory")
        self._add_line("#SBATCH -J grid_mp                 #job name")
        self._add_line("#SBATCH -N 1               ")
        self._add_line("#SBATCH --ntasks-per-node=1")
        self._add_line(f"#SBATCH --cpus-per-task={self._n_procs_to_use}")
        self._add_line(
            f"#SBATCH --time={str(self._hrs).zfill(2)}:{str(self._min).zfill(2)}:{str(self._sec).zfill(2)}"
        )
        self._add_line("#SBATCH --mail-type=ALL ")

        self._add_line(
            f"#SBATCH --mail-user={ronswanson_config.slurm.user_email}"
        )
        self._add_line("")

        self._add_line("module purge")

        if ronswanson_config.slurm.modules is not None:

            for m in ronswanson_config.slurm.modules:

                self._add_line(f"module load {m}")

        self._end_line()

        self._add_line("#add HDF5 library path to ld path")
        self._add_line("export LD_LIBRARY_PATH=$HDF5_HOME/lib:$LD_LIBRARY_PATH")

        self._add_line(
            f"srun {ronswanson_config.slurm.python} run_simulation.py ${{SLURM_ARRAY_TASK_ID}}"
        )


class SLURMGatherGenerator(ScriptGenerator):
    def __init__(
        self,
        file_name: str,
        n_procs: int,
        n_nodes: int,
        hrs: int,
        min: int,
        sec: int,
    ) -> None:

        self._n_procs: int = n_procs
        self._n_nodes: int = n_nodes

        self._hrs: int = hrs
        self._min: int = min
        self._sec: int = sec

        super().__init__(file_name)

    def _build_script(self) -> None:

        self._add_line("#!/bin/bash")
        self._add_line("")
        self._add_line("#SBATCH -o ./output/%A.out      #output file")
        self._add_line("#SBATCH -e ./output/%A.err      #error file")
        self._add_line("#SBATCH -D ./                      #working directory")
        self._add_line("#SBATCH -J gather_out                 #job name")
        self._add_line(f"#SBATCH -N {self._n_nodes}               ")
        self._add_line(f"#SBATCH --ntasks-per-node={self._n_procs}")

        self._add_line(
            f"#SBATCH --time={str(self._hrs).zfill(2)}:{str(self._min).zfill(2)}:{str(self._sec).zfill(2)}"
        )
        self._add_line("#SBATCH --mail-type=ALL ")

        self._add_line(
            f"#SBATCH --mail-user={ronswanson_config.slurm.user_email}"
        )
        self._add_line("")

        self._add_line("module purge")

        if ronswanson_config.slurm.mpi_modules is not None:

            for m in ronswanson_config.slurm.mpi_modules:

                self._add_line(f"module load {m}")

        self._end_line()

        self._add_line(
            f"srun {ronswanson_config.slurm.python} gather_results.py"
        )



import numpy as np

import ronswanson.database as database
from ronswanson.utils.logging import setup_logger

log = setup_logger(__name__)


def first_last_nonzero(arr, axis, invalid_val=-1):
    mask = arr > 0.0

    first = np.where(mask.any(axis=axis), mask.argmax(axis=axis), invalid_val)

    mask = np.flip(mask, axis=1)

    last = np.where(mask.any(axis=axis), mask.argmax(axis=axis), invalid_val)
    return first, -last


def check_for_non_contiguous_spectra(database: database.Database) -> np.ndarray:
    """
    find spectra with non-contiguous values (go to zero in between)

    :param database:
    :type database: Database
    :returns:

    """

    ok_idx = np.ones(database.n_entries, dtype=bool)

    first, last = first_last_nonzero(database.values, axis=1)

    for i, datum in enumerate(database.values):

        if np.any(datum[first[i] : last[i]] == 0):

            ok_idx[i] = 0

    log.info(f"there were {(~ok_idx).sum()} dirty entries")

    return ~ok_idx
